### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### This file holds various utility types and functions, especially for dealing with
### units and dates.
###

## Import and define units and dimensions
import Unitful: cm, m, km, ha, mg, g, kg, Length, Area, Mass
const m² = m^2
const km² = km^2

# enable division with different length/area unit types
Base.:(/)(x::S,y::T) where {S<:Length, T<:Length} = (upreferred(x)/m) / (upreferred(y)/m)
Base.:(/)(x::S,y::T) where {S<:Area, T<:Area} = (upreferred(x)/m²) / (upreferred(y)/m²)
Base.:(/)(x::S,y::T) where {S<:Mass, T<:Mass} = (upreferred(x)/g) / (upreferred(y)/g)

## Utility type and function for working wth recurring dates

"""
    AnnualDate

A type to handle recurring dates (e.g. migration, harvest).
Stores a month and a day, and can be compared against normal dates.
To save typing, a Tuple{Int64,Int64} is automatically converted to an
AnnualDate, allowing this syntax: `nestingend::AnnualDate = (August, 15)`.
"""
mutable struct AnnualDate
    month::Int64
    day::Int64

    # inner constructor to make sure inputs are in range
    AnnualDate(month::Int64, day::Int64) =
        if !(0 < month <= 12)
            Base.error("AnnualDate: month $month is out of range.") #TODO replace with exception
        elseif !(0 < day <= 31)
            # not strictly accurate (AnnualDate(February, 30) is possible), but good enough
            Base.error("AnnualDate: day $day is out of range.") #TODO replace with exception
        else
            new(month, day)
        end
end

# (automatically) convert integer tuples to AnnualDate.
# Allows writing `(August, 2)` instead of `AnnualDate(August, 2)` where an AnnualDate is expected.
AnnualDate(ad::Tuple{Int64,Int64}) = AnnualDate(ad...)
Base.convert(::Type{AnnualDate}, ad::Tuple{Int64,Int64}) = AnnualDate(ad)

# Interface with Dates
AnnualDate(date::Date) = AnnualDate(month(date), day(date))
Dates.month(ad::AnnualDate) = ad.month
Dates.day(ad::AnnualDate) = ad.day

# Instantiate a recurring date for a given year
"""
    thisyear(annualdate, model)
    nextyear(annualdate, model)
    lastyear(annualdate, model)

Convert an AnnualDate to a Date, using the current/next/previous year of the simulation run.
"""
thisyear(ad::AnnualDate, model::SimulationModel) = Date(year(model.date), ad.month, ad.day)
nextyear(ad::AnnualDate, model::SimulationModel) = Date(year(model.date)+1, ad.month, ad.day)
lastyear(ad::AnnualDate, model::SimulationModel) = Date(year(model.date)-1, ad.month, ad.day)

# Comparison between AnnualDates and with Dates
Base.:(==)(ad1::AnnualDate, ad2::AnnualDate) = (month(ad1) == month(ad2) && day(ad1) == day(ad2))
Base.:(==)(ad::AnnualDate, d::Date) = (month(d) == month(ad) && day(d) == day(ad))
Base.:(==)(d::Date, ad::AnnualDate) = (ad == d)
Base.:(<)(ad1::AnnualDate, ad2::AnnualDate) = (month(ad1) < month(ad2) ||
                                               (month(ad1) == month(ad2) && day(ad1) < day(ad2)))
Base.:(<)(ad::AnnualDate, d::Date) = (thisyear(ad) < d)
Base.:(<)(d::Date, ad::AnnualDate) = (d < thisyear(ad))

# Addition and subtraction of date periods
Base.:(+)(ad::AnnualDate, time::DatePeriod) = AnnualDate(thisyear(ad) + time)
Base.:(-)(ad::AnnualDate, time::DatePeriod) = AnnualDate(thisyear(ad) - time)
